/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.component.internal.multi;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;
import javax.inject.Singleton;

import org.xwiki.component.annotation.Component;
import org.xwiki.component.manager.ComponentLookupException;
import org.xwiki.component.manager.ComponentManager;

/**
 * Default implementation of {@link ComponentManagerManager}.
 * 
 * @version $Id: 99870654e787c46a9cf78d30e7d90d3282f67f54 $
 * @since 3.3M2
 */
@Component
@Singleton
public class DefaultComponentManagerManager implements ComponentManagerManager
{
    /**
     * The Factory to use to create Component Managers whenever a component is registered for a key that doesn't already
     * exist.
     */
    @Inject
    private ComponentManagerFactory defaultComponentManagerFactory;

    /**
     * The root component manager.
     */
    @Inject
    private ComponentManager rootComponentManager;

    /**
     * Holds Component Managers based on identifiers.
     */
    private Map<String, ComponentManager> componentManagers = new ConcurrentHashMap<String, ComponentManager>();

    @Override
    public ComponentManager getComponentManager(String path, boolean create)
    {
        ComponentManager componentManager;
        if (path == null) {
            componentManager = this.rootComponentManager;
        } else {
            componentManager = this.componentManagers.get(path);
            if (componentManager == null && create) {
                componentManager = createComponentManager(path);
                this.componentManagers.put(path, componentManager);
            }
        }

        return componentManager;
    }

    /**
     * Create a new {@link ComponentManager} for the provided id.
     * 
     * @param id the identifier of the component manager
     * @return a new {@link ComponentManager} instance
     */
    private ComponentManager createComponentManager(String id)
    {
        String prefix = getPrefix(id);

        ComponentManagerFactory componentManagerFactory;
        try {
            componentManagerFactory = this.rootComponentManager.getInstance(ComponentManagerFactory.class, prefix);
        } catch (ComponentLookupException e) {
            componentManagerFactory = this.defaultComponentManagerFactory;
        }

        return componentManagerFactory.createComponentManager(this.rootComponentManager);
    }

    /**
     * Extract prefix of the id used to find custom factory.
     * 
     * @param id the identifier of the component manager to create
     * @return the prefix of the id or null if none is provided
     */
    private String getPrefix(String id)
    {
        boolean escaped = false;
        StringBuilder typeBuilder = null;
        for (int i = 0; i < id.length(); ++i) {
            char c = id.charAt(i);
            if (escaped) {
                typeBuilder.append(c);
            } else {
                if (c == '\\') {
                    if (typeBuilder == null) {
                        typeBuilder = new StringBuilder();
                        typeBuilder.append(id.substring(0, i));
                    }
                } else if (c == ':') {
                    return id.substring(0, i);
                } else if (typeBuilder != null) {
                    typeBuilder.append(c);
                }
            }
        }

        return null;
    }
}
